package work.xuye.helper.checker;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;
import org.springframework.util.ObjectUtils;
import work.xuye.common.db.service.TaskManager;
import work.xuye.helper.properties.HelperProperties;
import work.xuye.helper.service.JdbcTemplateManager;

import java.util.*;

/**
 * @author xuye
 * @since 2023/7/13 14:57
 **/
@Slf4j
@Service
@RequiredArgsConstructor
public class DataDiffChecker {

    private final TaskManager taskManager;
    private final JdbcTemplateManager jdbcTemplateManager;
    private final HelperProperties helperProperties;


    public Map<String, List<Object>> checkDifferentKeys(Map<String, Object> map1, Map<String, Object> map2) {
        Map<String, List<Object>> map = new HashMap<>();
        for (String key : map1.keySet()) {
            if (map2.containsKey(key)) {
                Object value1 = map1.get(key);
                Object value2 = map2.get(key);

                if (!ObjectUtils.nullSafeEquals(value1, value2)) {
                    if (!helperProperties.getEqualsIgnoreField().contains(key)) {
                        map.put(key, List.of(value1, value2));
                        log.debug("@@@ key: {}, value1: {}, value2: {}", key, value1, value2);
                    }
                }
            } else {
                log.debug("@@@ key: {}, value1: {}, value2: {}", key, map1.get(key), null);
                ArrayList<Object> list = new ArrayList<>();
                list.add(map1.get(key));
                list.add(null);
                map.put(key, list);
            }
        }
        for (String key : map2.keySet()) {
            if (!map1.containsKey(key)) {
                log.debug("@@@ key: {}, value1: {}, value2: {}", key, null, map2.get(key));
                ArrayList<Object> list = new ArrayList<>();
                list.add(null);
                list.add(map2.get(key));
                map.put(key, list);
            }
        }
        return map;
    }


    public void check(List<Integer> ids, Integer intervalDays) {
        List<Integer> newsIds = helperProperties.getTaskTypeIdMap().get(NewsTypeEnum.NEWS);
        List<Integer> videoIds = helperProperties.getTaskTypeIdMap().get(NewsTypeEnum.VIDEO);
        if (ObjectUtils.isEmpty(ids)) {
            handleNews(newsIds, intervalDays);
//            handleVideo(videoIds);
        } else {
            for (Integer id : ids) {
                if (newsIds.contains(id)) {
                    handleNews(List.of(id), intervalDays);
                } else if (videoIds.contains(id)) {
//                    handleVideo(List.of(id));
                }
            }
        }
    }

    public void handleNews(List<Integer> ids, Integer intervalDays) {

        String source = helperProperties.getSourceDbName();
        String target = helperProperties.getTargetDbName();

        String sql = "select url from all_news where spider = ? and createDate > CURDATE() - INTERVAL ? DAY order by createDate desc";
        for (Integer id : ids) {
            String taskName = taskManager.getTaskInfoByTaskId(id).getTask().getName();
            log.info("-------------------------{}-------------------------", taskName);
            List<String> sourceUrl = jdbcTemplateManager.getTemplate(source).queryForList(sql, String.class, taskName, intervalDays);
            List<String> targetUrl = jdbcTemplateManager.getTemplate(target).queryForList(sql, String.class, taskName, intervalDays);

            // 判断交集数据
            List<String> intersection = new ArrayList<>(sourceUrl);
            intersection.retainAll(targetUrl);
            log.info("@@@ [{}] 记录数: [{}], [{}] 记录数: [{}], 交集条数: [{}]", source, sourceUrl.size(), target, targetUrl.size(), intersection.size());

            // 以 source 为基准，判断差集数据
            List<String> diff = findDifference(sourceUrl, intersection);
            if (diff.size() > 0) {
                log.info("@@@ [{}] 有，但是 [{}] 没有的记录数: {}", source, target, diff.size());
                checkDiffReason(diff, jdbcTemplateManager.getTemplate(target));
            }


            // 以 target 为基准，判断差集数据
            diff = findDifference(targetUrl, intersection);
            if (diff.size() > 0) {
                log.info("@@@ [{}] 有，但是 [{}] 没有的记录数: {}", target, source, diff.size());
                checkDiffReason(diff, jdbcTemplateManager.getTemplate(source));
            }


            // 判断交集数据是否相同
            findFieldsDiff(intersection, jdbcTemplateManager.getTemplate(source), jdbcTemplateManager.getTemplate(target));
        }
    }

    private void findFieldsDiff(List<String> urls, JdbcTemplate jdbcTemplate, JdbcTemplate jdbcTemplate2) {
        int count = 0;
        Map<String, Map<String, List<Object>>> result = new HashMap<>();
        String sql = "select * from all_news where url = ? limit 1";
        for (String s : urls) {
            if (urls.size() > helperProperties.getThreshold() && Math.random() > helperProperties.getSampleRate()) {
                continue;
            }
            count++;
            Map<String, Object> map = jdbcTemplate.queryForMap(sql, s);
            Map<String, Object> map2 = jdbcTemplate2.queryForMap(sql, s);
            Map<String, List<Object>> differentKeys = checkDifferentKeys(map, map2);
            if (differentKeys.size() > 0) {
                result.put(s, differentKeys);
            }
        }
        if (ObjectUtils.isEmpty(result)) {
            log.info("@@@ 在 [{}] 条中抽检了 [{}] 条，没有字段差异", urls.size(), count);
        } else {

            Set<String> keys = new HashSet<>();
            for (Map<String, List<Object>> value : result.values()) {
                keys.addAll(value.keySet());
            }
            log.warn("@@@ 共在 [{}] 条中抽检了 [{}] 条，有 [{}] 条有字段差异, 差异字段: {}", urls.size(), count, result.size(), keys);
        }
    }

    public List<String> findDifference(List<String> source, List<String> intersection) {
        List<String> diff = new ArrayList<>(source);
        diff.removeAll(intersection);
        return diff;
    }

    public void checkDiffReason(List<String> urls, JdbcTemplate jdbcTemplate) {
        List<String> occupied = new ArrayList<>();
        List<String> notExist = new ArrayList<>();

        String sql = "select spider from all_news where url = ?";
        for (String url : urls) {
            List<String> spiders = jdbcTemplate.queryForList(sql, String.class, url);
            if (spiders.size() == 1) {
                occupied.add(url);
            } else {
                notExist.add(url);
            }
        }

        if (!ObjectUtils.isEmpty(occupied)) {
            log.info("@@ 已经被其他爬虫提前爬取的记录条数`: {}", occupied.size());
        }
        if (!ObjectUtils.isEmpty(notExist)) {
            if (notExist.size() > 5) {
                log.error("@@ ⚠️不存在的记录条数: {}", notExist.size());
            } else {
                log.warn("@@ 不存在的记录条数: {}, url: {}", notExist.size(), notExist);
            }
        }


    }


}
